#!/usr/bin/python
# -*- coding: utf-8 -*-

# This file is part of bzz.
# https://github.com/heynemann/bzz

# Licensed under the MIT license:
# http://www.opensource.org/licenses/MIT-license
# Copyright (c) 2014 Bernardo Heynemann heynemann@gmail.com


from tornado.concurrent import is_future
import tornado.gen as gen
import blinker


class Signal(blinker.NamedSignal):
    @gen.coroutine
    def send(self, *sender, **kwargs):
        if len(sender) == 0:
            sender = None
        elif len(sender) > 1:
            raise TypeError('send() accepts only one positional argument, '
                            '%s given' % len(sender))
        else:
            sender = sender[0]

        if not self.receivers:
            raise gen.Return([])

        results = []
        for receiver in self.receivers_for(sender):
            result = receiver(sender, **kwargs)

            if is_future(result):
                result = yield result
            results.append((receiver, result))

        raise gen.Return(results)


class Namespace(dict):
    """A mapping of signal names to signals."""

    def signal(self, name, doc=None):
        """Return the :class:`NamedSignal` *name*, creating it if required.

        Repeated calls to this function will return the same signal object.

        """
        try:
            return self[name]
        except KeyError:
            return self.setdefault(name, Signal(name, doc))


signal = Namespace().signal

pre_get_instance = signal('bzz.pre-get-instance')
post_get_instance = signal('bzz.post-get-instance')

pre_get_list = signal('bzz.pre-get-list')
post_get_list = signal('bzz.post-get-list')

pre_create_instance = signal('bzz.pre-create-instance')
post_create_instance = signal('bzz.post-create-instance')

pre_update_instance = signal('bzz.pre-update-instance')
post_update_instance = signal('bzz.post-update-instance')

pre_delete_instance = signal('bzz.pre-delete-instance')
post_delete_instance = signal('bzz.post-delete-instance')

authorized_user = signal('bzz.authorized-user')
unauthorized_user = signal('bzz.unauthorized-user')

pre_get_user_details = signal('bzz.pre-get-user-details')



#!/usr/bin/python
# -*- coding: utf-8 -*-

# This file is part of bzz.
# https://github.com/heynemann/bzz

# Licensed under the MIT license:
# http://www.opensource.org/licenses/MIT-license
# Copyright (c) 2014 Bernardo Heynemann heynemann@gmail.com

import tornado.web
import tornado.gen as gen
from six.moves.urllib.parse import unquote

import bzz.core as core
import bzz.signals as signals
import bzz.utils as utils


AVAILABLE_PROVIDERS = {
    'mongoengine': 'bzz.providers.mongoengine_provider.MongoEngineProvider'
}


class ModelHive(object):
    @classmethod
    def routes_for(cls, provider, model, prefix='', resource_name=None):
        '''
        Returns the list of routes for the specified model.

        * [POST] Create new instances;
        * [PUT] Update existing instances;
        * [DELETE] Delete existing instances;
        * [GET] Retrieve existing instances with the id for the instance;
        * [GET] List existing instances (and filter them);

        :param provider: The ORM provider to be used for this model
        :type provider: Full-name provider or built-in provider
        :param model: The model to be mapped
        :type model: Class Type
        :param prefix: Optional argument to include a prefix route (i.e.: '/api');
        :type prefix: string
        :param resource_name: an optional argument that can be specified to change the route name. If no resource_name specified the route name is the __class__.__name__ for the specified model with underscores instead of camel case.
        :type resource_name: string
        :returns: route list (can be flattened with bzz.flatten)

        If you specify a prefix of '/api/' as well as resource_name of 'people' your route would be similar to:

        http://myserver/api/people/ (do a post to this url to create a new person)

        Usage:

        .. testcode:: model_hive_example_1

           import tornado.web
           from mongoengine import *
           import bzz

           server = None

           # just create your own documents
           class User(Document):
              __collection__ = "MongoEngineHandlerUser"
              name = StringField()

           def create_user():
              # let's create a new user by posting it's data
              http_client.fetch(
                 'http://localhost:8890/user/',
                 method='POST',
                 body='name=Bernardo%20Heynemann',
                 callback=handle_user_created
              )

           def handle_user_created(response):
              # just making sure we got the actual user
              try:
                 assert response.code == 200, response.code
              finally:
                 io_loop.stop()

           # bzz includes a helper to return the routes for your models
           # returns a list of routes that match '/user/<user-id>/' and allows for:
           routes = bzz.ModelHive.routes_for('mongoengine', User)

           User.objects.delete()
           application = tornado.web.Application(routes)
           server = HTTPServer(application, io_loop=io_loop)
           server.listen(8895)
           io_loop.add_timeout(1, create_user)
           io_loop.start()
        '''
        provider_name = AVAILABLE_PROVIDERS.get(provider, provider)
        provider_class = utils.get_class(provider_name)
        name = resource_name
        if name is None:
            name = utils.convert(model.__name__)

        details_regex = r'/(%s(?:/[^/]+)?)((?:/[^/]+)*)/?'

        details_regex = utils.add_prefix(prefix, details_regex)

        tree = provider_class.get_tree(model)

        options = dict(model=model, name=name, prefix=prefix, tree=tree)
        routes = core.RouteList()

        routes.append(
            (details_regex % name, provider_class, options)
        )

        return routes


class ModelProvider(tornado.web.RequestHandler):
    @classmethod
    def get_tree(cls, model, node=None):
        if node is None:
            node = core.Node(cls.get_model_name(model), is_root=True)

        node.target_name = cls.get_model_collection(model)
        node.is_multiple = False

        if node.target_name is None:
            node.target_name = node.slug

        node.model_type = model

        cls.parse_children(model, node.children)

        return node

    @classmethod
    def parse_children(cls, model, collection):
        for field_name, field in cls.get_model_fields(model).items():
            child_node = core.Node(field_name)
            collection[field_name] = child_node

            child_node.is_multiple = cls.is_list_field(field)
            child_node.target_name = cls.get_field_target_name(field)
            child_node.allows_create_on_associate = \
                cls.allows_create_on_associate(field)
            child_node.is_lazy_loaded = \
                cls.is_lazy_loaded(field)

            child_node.model_type = cls.get_model(field)

            if child_node.model_type is not None:
                cls.parse_children(child_node.model_type, child_node.children)

    def get_node(self, path):
        if '.' not in path:
            return self.tree.get(path, None)

        node = self.tree
        for item in path.split('.'):
            node = node.get(item, None)
            if node is None:
                return None

        return node

    def initialize(self, model, name, prefix, tree):
        self.model = model
        self.name = name
        self.prefix = prefix
        self.tree = tree

    def write_json(self, obj):
        self.set_header("Content-Type", "application/json")
        self.write(utils.dumps(obj))

    def parse_arguments(self, args):
        args = [arg.lstrip('/') for arg in args if arg]
        if len(args) == 1:
            return args

        parts = args[1].split('/')

        items = []
        current_item = []
        for part in parts:
            current_item.append(part)
            if len(current_item) == 2:
                items.append('/'.join(current_item))
                current_item = []
        if current_item:
            items.append(current_item[0])

        return [args[0]] + items

    @classmethod
    def get_path_from_args(cls, args):
        path = ".".join([arg.split('/')[0] for arg in args[1:]])
        return path

    @gen.coroutine
    def get(self, *args, **kwargs):
        args = self.parse_arguments(args)
        path = self.get_path_from_args(args)
        node = self.tree.find_by_path(path)

        if (node.is_root or node.is_multiple) and '/' not in args[-1]:
            yield signals.pre_get_list.send(node.model_type, arguments=args, handler=self)
            yield self.handle_get_list(args)
        else:
            yield signals.pre_get_instance.send(node.model_type, arguments=args, handler=self)
            yield self.handle_get_one(args)

    @gen.coroutine
    def handle_get_one(self, args):
        success, obj, parent = yield self.get_instance_from_args(args)
        if not success:
            return

        if obj is None:
            self.send_error(status_code=404)
            return

        yield signals.post_get_instance.send(obj.__class__, instance=obj, handler=self)

        self.write_json(self.dump_instance(obj))
        self.finish()

    @gen.coroutine
    def handle_get_list(self, args):
        if '/' not in args[0] and len(args) > 1:
            # /team/user -> missing user pk
            self.send_error(status_code=400)
            return

        if len(args) == 1:
            request_data = self.get_request_data()
            try:
                page = int(request_data.pop('page', 1))
            except ValueError:
                page = 1

            try:
                per_page = int(request_data.pop('per_page', 20))
            except ValueError:
                per_page = 20

            items = yield self.get_list(page=page, per_page=per_page, filters=request_data)
            model_type = self.model
        else:
            success, items, parent = yield self.get_instance_from_args(args)
            if not success:
                self.send_error(status_code=400)
                return

            model_type = self.get_property_model(parent, args[-1])

        yield signals.post_get_list.send(model_type, items=items, handler=self)

        self.write_json(self.dump_list(items))
        self.finish()

    @gen.coroutine
    def get_instance_from_args(self, args):
        model, pk = args[0].split('/')
        obj = yield self.get_instance(pk)

        if len(args) == 1:
            raise gen.Return((True, obj, None))

        obj, parent = yield self.get_instance_property(obj, args[1:])

        if obj is None:
            self.send_error(status_code=404)
            raise gen.Return((False, None))

        raise gen.Return((True, obj, parent))

    @gen.coroutine
    def get_instance_property(self, obj, path):
        parts = [part.lstrip('/').split('/') for part in path if part]
        parent = obj
        for part in parts:
            path = part[0]
            pk = None

            if len(part) > 1:
                pk = part[1]

            obj = getattr(parent, path)

            if pk is not None:
                if isinstance(obj, (list, tuple)):
                    for item in obj:
                        instance_id = yield self.get_instance_id(item)
                        if instance_id == pk:
                            obj = item
                        if part != parts[-1]:
                            parent = obj

        raise gen.Return([obj, parent])

    @gen.coroutine
    def post(self, *args, **kwargs):
        args = self.parse_arguments(args)
        model_type = yield self.get_model_from_path(args)

        yield signals.pre_create_instance.send(
            model_type,
            arguments=args,
            handler=self,
        )

        instance = None

        if len(args) == 1:
            if '/' in args[0]:
                self.send_error(400)
                return

            instance, error = yield self.handle_create_one(args)
        else:
            is_reference = yield self.is_reference(args)
            if is_reference:
                instance, error = yield self.handle_find_and_associate(args)
            else:
                instance, error = yield self.handle_create_and_associate(args)

        if error is not None:
            status_code, error = error
            self.set_status(status_code)
            self.write(str(error))
            return

        yield signals.post_create_instance.send(
            instance.__class__,
            instance=instance,
            handler=self
        )
        pk = yield self.get_instance_id(instance)
        self.set_header('X-Created-Id', pk)
        self.set_header('location', '/%s%s/%s/' % (
            self.prefix,
            self.name,
            pk
        ))
        self.write('OK')

    @gen.coroutine
    def handle_create_one(self, args):
        instance, error = yield self.save_new_instance(self.model, self.get_request_data())
        raise gen.Return((instance, error))

    @gen.coroutine
    def handle_create_and_associate(self, args):
        path, pk = args[0].split('/')
        root = yield self.get_instance(pk)
        model_type = self.get_model_type(root, args[1:])
        instance, error = yield self.save_new_instance(model_type, self.get_request_data())
        if error is not None:
            raise gen.Return((None, error))

        _, error = yield self.associate_instance(root, args[-1], instance)
        if error is not None:
            raise gen.Return((None, error))

        raise gen.Return((instance, error))

    @gen.coroutine
    def handle_find_and_associate(self, args):
        path, pk = args[0].split('/')
        root = yield self.get_instance(pk)

        _, parent = yield self.get_instance_property(root, args[1:])
        request_data = self.get_request_data()
        model_type = self.get_property_model(parent, args[-1])
        key = "%s[]" % args[-1]
        value = request_data[key]
        instance = yield self.get_instance(value, model=model_type)

        _, error = yield self.associate_instance(root, args[-1], instance)
        if error is not None:
            raise gen.Return((None, error))

        raise gen.Return((instance, None))

    def get_model_type(self, obj, args):
        for index, arg in enumerate(args[:-1]):
            property_name, pk = arg.split('/')
            obj = getattr(obj, property_name)

        return self.get_property_model(obj, args[-1])

    @gen.coroutine
    def put(self, *args, **kwargs):
        args = self.parse_arguments(args)
        model_type = yield self.get_model_from_path(args)

        yield signals.pre_update_instance.send(
            model_type,
            arguments=args,
            handler=self
        )

        instance = None

        if len(args) == 1 and '/' not in args[0]:
            self.send_error(400)
            return

        is_multiple = yield self.is_multiple(args)
        is_reference = yield self.is_reference(args)
        if is_multiple and is_reference:
            self.send_error(400)
            return

        instance, updated, model = yield self.handle_update(args)
        yield signals.post_update_instance.send(model, instance=instance, updated_fields=updated, handler=self)
        self.write('OK')

    @gen.coroutine
    def handle_update(self, args):
        path, pk = args[0].split('/')
        root = yield self.get_instance(pk)
        model_type = root.__class__
        instance = parent = None

        if not self.validate_update_request_data(root, model_type):
            self.send_error(400, reason="Invalid multiple field")
            raise tornado.web.Finish()

        if len(args) > 1:
            instance, parent = yield self.get_instance_property(root, args[1:])
            model_type = instance.__class__
            property_name, pk = args[-1].split('/')
        error, instance, updated = yield self.update_instance(pk, self.get_request_data(), model_type, instance, parent)

        if error is not None:
            status_code, error = error
            self.set_status(status_code)
            self.write(str(error))
            return

        raise gen.Return([instance, updated, model_type])

    def validate_update_request_data(self, root, model_type):
        data = self.get_request_data()

        for key, value in data.items():
            if key.endswith('[]'):
                return False

        return True

    @gen.coroutine
    def delete(self, *args, **kwargs):
        args = self.parse_arguments(args)

        if len(args) == 1 and '/' not in args[0]:
            self.send_error(400)
            return

        model_type = yield self.get_model_from_path(args)
        yield signals.pre_delete_instance.send(model_type, arguments=args, handler=self)

        path, pk = args[0].split('/')
        root = yield self.get_instance(pk)
        instance = None

        if len(args) > 1:
            instance, parent = yield self.get_instance_property(root, args[1:])
            property_name, pk = args[-1], None
            if '/' in property_name:
                property_name, pk = property_name.split('/')
            instance, error = yield self.handle_delete_association(parent, instance, property_name)
        else:
            instance = yield self.handle_delete_instance(pk)
            error = None

        if error is not None:
            status_code, error = error
            self.set_status(status_code)
            self.write(str(error))
            return

        if instance:
            yield signals.post_delete_instance.send(model_type, instance=instance, handler=self)
            self.write('OK')
        else:
            self.write('FAIL')

    @gen.coroutine
    def handle_delete_instance(self, pk):
        instance = yield self.delete_instance(pk)
        raise gen.Return(instance)

    @gen.coroutine
    def handle_delete_association(self, parent, instance, property_name):
        field = parent._fields.get(property_name)
        if self.is_list_field(field):
            property_list = getattr(parent, property_name, [])

            try:
                property_list.remove(instance)
            except ValueError:
                self.send_error(400)
                return
        else:
            setattr(parent, property_name, None)

        _, error = yield self.save_instance(parent)

        raise gen.Return((instance, error))

    @gen.coroutine
    def list(self):
        items = yield self.get_list()
        dump = []
        for item in items:
            dump.append(self.dump_object(item))

        self.write_json(dump)

    def get_request_data(self):
        data = {}

        if self.request.body:
            items = self.request.body.decode('utf-8').split('&')
            for item in items:
                if '=' in item:
                    key, value = item.split('=')
                else:
                    key, value = 'item', item

                if key in data:
                    if not isinstance(data[key], (tuple, list)):
                        old = data[key]
                        data[key] = []
                        data[key].append(old)
                    data[key].append(unquote(value))
                else:
                    data[key] = unquote(value)
        else:
            for arg in list(self.request.arguments.keys()):
                data[arg] = self.get_argument(arg)
                if data[arg] == '':  # Tornado 3.0+ compatibility... Hard to test...
                    data[arg] = None

        return data

    def dump_object(self, instance):
        return utils.dumps(instance)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# This file is part of bzz.
# https://github.com/heynemann/bzz

# Licensed under the MIT license:
# http://www.opensource.org/licenses/MIT-license
# Copyright (c) 2014 Bernardo Heynemann heynemann@gmail.com

import tornado.web
from collections import OrderedDict

import bzz.core as core


class MockedRoutesHandler(tornado.web.RequestHandler):
    def __init__(self, *args, **kwargs):
        self.handler_methods = kwargs.pop('handler_methods')
        super(MockedRoutesHandler, self).__init__(*args, **kwargs)

    def prepare(self):
        if self.request.method in self.handler_methods or '*' in self.handler_methods:
            response = self.handler_methods.get(self.request.method, None) or self.handler_methods.get('*')
            status = response.get('status', 200)
            body = response.get('body', '')

            if hasattr(body, '__call__'):
                body = body(self.request)

            if status >= 400:
                raise tornado.web.HTTPError(status, body)

            if 'cookies' in response:
                for cookie, value in response['cookies'].items():
                    self.set_cookie(cookie, value)

            self.write(body)
            self.set_status(status)
            self.finish()


class MockHive(object):
    @classmethod
    def routes_for(cls, routes_tuple):
        """
        Returns a tuples list of paths, tornado ready
        """
        routes = OrderedDict()

        for route in routes_tuple:
            if not route[1] in routes:
                routes[route[1]] = {}
            if not route[0] in routes[route[1]]:
                routes[route[1]][route[0]] = {}
            if len(route) > 2:
                result = route[2]
            else:
                result = {'body': '', 'status': 200}

            routes[route[1]][route[0]] = result

        result = core.RouteList()
        for route, methods in routes.items():
            route_tuple = (route, MockedRoutesHandler, dict(handler_methods=methods))
            result.append(route_tuple)

        return result



#!/usr/bin/python
# -*- coding: utf-8 -*-

# This file is part of bzz.
# https://github.com/heynemann/bzz

# Licensed under the MIT license:
# http://www.opensource.org/licenses/MIT-license
# Copyright (c) 2014 Bernardo Heynemann heynemann@gmail.com

import logging

import tornado.gen as gen
from tornado import httpclient

from bzz.auth import AuthProvider
import bzz.utils as utils


class GoogleProvider(AuthProvider):
    '''
    Provider to perform authentication with Google OAUTH Apis.
    '''
    API_URL = 'https://www.googleapis.com/oauth2/v1/userinfo?access_token={}'

    @gen.coroutine
    def authenticate(self, access_token, proxy_info=None):
        '''
        Try to get Google user info and returns it if
        the given access_token get`s a valid user info in a string
        json format. If the response was not an status code 200 or
        get an error on Json, None was returned.

        Example of return on success::

            {
                id: "1234567890abcdef",
                email: "...@gmail.com",
                name: "Ricardo L. Dani",
                provider: "google"
            }
        '''

        response = yield self._fetch_userinfo(access_token, proxy_info)

        if response.code == 200:
            body = utils.loads(response.body)
            if not body.get('error'):
                raise gen.Return({
                    'email': body.get("email"),
                    'name': body.get("name"),
                    'id': body.get("id"),
                    'provider': self.get_name()
                })

        raise gen.Return(None)

    @gen.coroutine
    def _fetch_userinfo(self, access_token, proxy_info):
        url = self.API_URL.format(access_token)
        logging.info('Requesting %s with proxy %s...' % (url, proxy_info))
        req = httpclient.HTTPRequest(url, **proxy_info) if proxy_info else url
        try:
            response = yield self.http_client.fetch(req)
        except httpclient.HTTPError as e:
            response = e.response
        raise gen.Return(response)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# This file is part of bzz.
# https://github.com/heynemann/bzz

# Licensed under the MIT license:
# http://www.opensource.org/licenses/MIT-license
# Copyright (c) 2014 Bernardo Heynemann heynemann@gmail.com

import sys
import math

import tornado.gen as gen
import mongoengine

import bzz.model as bzz
import bzz.utils as utils


class MongoEngineProvider(bzz.ModelProvider):
    @classmethod
    def get_model_name(cls, model):
        return model.__name__

    @classmethod
    def get_model_collection(cls, model):
        return model._meta.get('collection', None)

    @classmethod
    def get_model_fields(cls, model):
        return model._fields

    @classmethod
    def get_model(cls, field):
        return cls.get_document_type(field)

    @classmethod
    def get_field_target_name(cls, field):
        return field.db_field

    @classmethod
    def get_document_type(cls, field):
        if cls.is_list_field(field):
            field = field.field
        return getattr(field, 'document_type', None)

    @classmethod
    def allows_create_on_associate(cls, field):
        if cls.is_list_field(field):
            field = field.field

        return cls.is_embedded_field(field)

    @classmethod
    def is_lazy_loaded(cls, field):
        if cls.is_list_field(field):
            field = field.field

        return cls.is_reference_field(field)

    @classmethod
    def is_list_field(cls, field):
        return isinstance(field, mongoengine.ListField)

    @classmethod
    def is_reference_field(cls, field):
        return isinstance(field, mongoengine.ReferenceField)

    @classmethod
    def is_embedded_field(cls, field):
        return isinstance(field, mongoengine.EmbeddedDocumentField)

    @gen.coroutine
    def save_new_instance(self, model, data):
        instance = model()

        for key, value in data.items():
            if '.' in key or '[]' in key:
                yield self.fill_property(model, instance, key, value)
            else:
                field = instance._fields.get(key)
                if self.is_reference_field(field):
                    value = yield self.get_instance(
                        value,
                        self.get_model(field)
                    )
                setattr(instance, key, value)

        if isinstance(instance, mongoengine.Document):
            try:
                instance.save()
            except mongoengine.NotUniqueError:
                err = sys.exc_info()[1]
                raise gen.Return((None, (409, err)))
            except mongoengine.ValidationError:
                err = sys.exc_info()[1]
                raise gen.Return((None, (400, err)))

        raise gen.Return((instance, None))

    @gen.coroutine
    def fill_property(self, model, instance, key, value, updated_fields=None):
        parts = key.split('.')
        field_name = parts[0]
        multiple = False
        if field_name.endswith('[]'):
            multiple = True
            field_name = field_name.replace('[]', '')

        property_name = '.'.join(parts[1:])

        if getattr(instance, field_name, None) is None:
            field = model._fields[field_name]
            embedded_document = field.document_type()

            if updated_fields is not None:
                updated_fields[field_name] = {
                    'from': getattr(instance, field_name),
                    'to': value
                }

            setattr(instance, field_name, embedded_document)

        if '.' not in property_name:
            if updated_fields is not None:
                updated_fields[field_name] = {
                    'from': getattr(instance, field_name),
                    'to': str(value)
                }

            field = getattr(model, field_name, None)
            child_model = self.get_model(field)
            if multiple and self.is_list_field(field):
                if not isinstance(value, (tuple, list)):
                    value = [value]

                list_property = getattr(instance, field_name)
                for item in value:
                    child_instance = yield self.get_instance(
                        item, model=child_model
                    )
                    list_property.append(child_instance)
            else:
                setattr(getattr(instance, field_name), property_name, value)
        else:
            new_instance = getattr(instance, field_name)
            yield self.fill_property(
                new_instance.__class__, new_instance,
                property_name, value
            )

    @gen.coroutine
    def update_instance(
            self, pk, data, model=None, instance=None, parent=None):
        if model is None:
            model = self.model

        if instance is None:
            instance = yield self.get_instance(pk, model)

        updated_fields = {}
        for field_name, value in self.get_request_data().items():
            if '.' in field_name:
                yield self.fill_property(
                    model, instance, field_name, value, updated_fields
                )
            else:
                field = instance._fields.get(field_name)
                if self.is_reference_field(field):
                    value = yield self.get_instance(
                        value, self.get_model(field)
                    )
                updated_fields[field_name] = {
                    'from': getattr(instance, field_name),
                    'to': value
                }
                setattr(instance, field_name, value)

        if parent and isinstance(instance, mongoengine.EmbeddedDocument):
            _, error = yield self.save_instance(parent)
        else:
            _, error = yield self.save_instance(instance)

        raise gen.Return((error, instance, updated_fields))

    @gen.coroutine
    def save_instance(self, instance):
        error = None
        try:
            instance.save()
        except mongoengine.NotUniqueError:
            err = sys.exc_info()[1]
            raise gen.Return((None, (409, err)))
        except mongoengine.ValidationError:
            err = sys.exc_info()[1]
            raise gen.Return((None, (400, err)))

        raise gen.Return((instance, error))

    @gen.coroutine
    def delete_instance(self, pk):
        instance = yield self.get_instance(pk)
        if instance is not None:
            instance.delete()
        raise gen.Return(instance)

    @gen.coroutine
    def get_instance(self, instance_id, model=None):
        if model is None:
            model = self.model

        queryset = model.objects
        if hasattr(model, 'get_instance_queryset'):
            queryset = model.get_instance_queryset(model, queryset, instance_id, self)

        instance = None
        field = self.get_id_field_name(model)

        if instance_id:
            instance = queryset.filter(**{field: instance_id}).first()

        raise gen.Return(instance)

    @gen.coroutine
    def get_list(self, items=None, page=1, per_page=20, filters=None):
        if filters is None:
            filters = {}

        queryset = self.model.objects
        if hasattr(self.model, 'get_list_queryset'):
            queryset = self.model.get_list_queryset(queryset, self)

        if filters:
            queryset = queryset.filter(**dict([
                (key.replace('.', '__'), value)
                for key, value in filters.items()
            ]))

        pages = int(math.ceil(queryset.count() / float(per_page)))
        if pages == 0:
            raise gen.Return([])

        if page > pages:
            page = pages

        page -= 1

        start = per_page * page
        stop = start + per_page

        items = queryset.all()[start:stop]
        raise gen.Return(items)

    def dump_list(self, items):
        dumped = []

        for item in items:
            dumped.append(self.dump_instance(item))

        return dumped

    def dump_instance(self, instance):
        if instance is None:
            return {}

        method = getattr(instance, 'to_dict', None)

        if method:
            return method()

        return utils.loads(instance.to_json())

    @gen.coroutine
    def get_instance_id(self, instance):
        field = getattr(instance.__class__, 'get_id_field_name', None)
        if field:
            raise gen.Return(str(getattr(instance, field().name)))

        raise gen.Return(str(instance.id))

    def get_id_field_name(self, model=None):
        if model is None:
            model = self.model
        field = getattr(model, 'get_id_field_name', None)
        if field:
            return field().name

        return 'id'

    @gen.coroutine
    def associate_instance(self, obj, field_name, instance):
        if obj is None:
            return

        field = obj._fields.get(field_name)
        if self.is_list_field(field):
            getattr(obj, field_name).append(instance)
        else:
            setattr(obj, field_name, instance)

        try:
            obj.save()
        except mongoengine.NotUniqueError:
            err = sys.exc_info()[1]
            raise gen.Return((None, (409, err)))
        except mongoengine.ValidationError:
            err = sys.exc_info()[1]
            raise gen.Return((None, (400, err)))

        raise gen.Return((obj, None))

    def get_property_model(self, obj, field_name):
        property_name = field_name
        pk = None

        if '/' in field_name:
            property_name, pk = field_name.split('/')

        field = obj._fields[property_name]
        return self.get_document_type(field)

    @gen.coroutine
    def is_multiple(self, path):
        parts = [part.lstrip('/').split('/') for part in path if part]
        to_return = False
        model = self.model

        if len(parts) == 1 and len(parts[0]) == 1:
            raise gen.Return(True)

        for part in parts[1:]:
            to_return = False
            path = part[0]
            field = getattr(model, path)
            to_return = self.is_list_field(field)
            model = self.get_model(field)

        raise gen.Return(to_return)

    @gen.coroutine
    def is_reference(self, path):
        parts = [part.lstrip('/').split('/') for part in path if part]
        to_return = False
        model = self.model

        for part in parts[1:]:
            to_return = False
            model_path = part[0]
            field = getattr(model, model_path)

            if self.is_list_field(field):
                field = field.field

            to_return = self.is_reference_field(field)
            model = self.get_model(field)

        raise gen.Return(to_return)

    @gen.coroutine
    def get_model_from_path(self, path):
        parts = [part.lstrip('/').split('/') for part in path if part]
        model = self.model

        if len(parts) == 1 and len(parts[0]) == 1:
            raise gen.Return(model)

        for part in parts[1:]:
            path = part[0]
            field = getattr(model, path)
            model = self.get_model(field)

        raise gen.Return(model)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# This file is part of bzz.
# https://github.com/heynemann/bzz

# Licensed under the MIT license:
# http://www.opensource.org/licenses/MIT-license
# Copyright (c) 2014 Bernardo Heynemann heynemann@gmail.com

import slugify


class RouteList(list):
    pass


class Node(object):
    def __init__(self, name, is_root=False):
        if not name:
            raise ValueError("Can't create unnamed node.")

        self.is_root = is_root
        self.name = name
        self.slug = slugify.slugify(self.name.lower())
        self.target_name = name
        self.model_type = None
        self.is_multiple = False
        self.allows_create_on_associate = False
        self.lazy_loaded = False
        self.children = {}
        self.required_children = []

    def find_by_path(self, path):
        if not path:
            return self
        if '.' not in path:
            return self.children.get(path, None)

        obj = self
        for part in path.split('.'):
            obj = obj.children.get(part, None)

            if obj is None:
                break

        return obj
